<?php

class RealTime {

	const GUI_VIEWBY = 'realtime_viewby';
	const GUI_SORTBY = 'licensestatus_sortby';
	const GUI_SORTASC = 'licensestatus_sortasc';

	const GUI_USER_SORTBY = 'licensestatus_user_sortby';
	const GUI_USER_SORTASC = 'licensestatus_user_sortasc';

	const GUI_AUTO_RELOAD = 'licensestatus_autoreload';

	const GUI_USERBLOCK = 'licensestatus_userblock';

	const GUI_PRODUCT_MAPPING = 'licensestatus_productmapping';

	const GUI_TIMEZONE_OFFSET = 'licensestatus_timezoneoffset';
	
	const GUI_CLIENT_TIME = 'licensestatus_clienttime';
	const MAX = 'licensestatus_max';

	const PRODUCT = "product";
	const FEATURE = "feature";
	const USER = "user";

	static private $instance;
	private $realTimeMap;
	private $userData;
	private $collectionTime;
	private $URL;

	private function __construct() {

		
		// Add products if they are selected
		if (count($_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"]) != 0 && count($_SESSION["REALTIME"]["SELECTION"]["FEATURES"]) != 0) {
			//$URL .= "?products=" . implode(":", $_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"]);
		}

		$data = file_get_contents(self::getUrl());
		// TODO: SHOULD BE FIXED IN BACKEND!
		$data = str_replace("&", "&amp;", $data);
		
		

		// If we have SimpleXML, we use that, since its a LOT faster..
		if (class_exists('SimpleXMLElement', false)) {
			try {
				$this->realTimeMap = $this->readXML($data);
			} catch (Exception $e) {
				$noData = true;
			}
		}

		// - no SimpleXML php, using old...
		else {
			try {
				$this->realTimeMap = $this->readXML_OLD($data);
			} catch (Exception $e) {
				$noData = true;
			}
		}

		// Show no data page...
		if (is_null($this->realTimeMap) || isset($noData)) {
			header("Location: nodata.php");
			exit;
		}
		
		
		/*$this->populateUserData();*/


	}

	static private function getUrl() {

		// If we have a static file in classes, we use this as static information
		if (file_exists(ServerConfig::getConfigPath() . "/../../helpers/license_status.xml")) {
			return ServerConfig::getConfigPath() . "/../../helpers/license_status.xml";
		}

		// return url to the license status xml (from backend)
		/*$chars = ServerConfig::getEncoding() == "SJIS" ? "shift_jis" : null ;
		if($chars){
			return ServerConfig::getServerUrl() . 'license_status.pl?encoding='.$chars;
		}else */
		
		$encoding = ServerConfig::getCharset();
		return ServerConfig::getServerUrl() . 'license_status.pl?encoding=' . $encoding;
	}

	//get the longest substring match between two strings
	function strlcs($str1, $str2){
		$m = strlen($str1);
		$n = strlen($str2);
		$L = array();
		$z = 0;
		$ret = array();

		for($i=0; $i<$m; $i++){
			$L[$i] = array();
			for($j=0; $j<$n; $j++){
				$L[$i][$j] = 0;
			}
		}

		for($i=0; $i<$m; $i++){
			for($j=0; $j<$n; $j++){
				if( $str1[$i] == $str2[$j] ){
					$L[$i][$j] = $L[$i-1][$j-1] + 1;
					if( $L[$i][$j] > $z ){
						$z = $L[$i][$j];
						$ret = array();
					}
					if( $L[$i][$j] == $z )
					$ret[] = substr($str1, $i-$z+1, $z);
				}
			}
		}
		return count($ret);
	}


	//format the license path
	private function formatLicense($license){
		$license = str_replace("\\", "\\\\", $license);
		return $license;
	}
		
	static public function isRealTimeEnabled() {
		return ServerConfig::getLicenseStatus() ? true : false;
		/*
		 if (isset($_SESSION["REALTIMEENABLED"])) {
			return $_SESSION["REALTIMEENABLED"];
			}
			$data = file_get_contents(self::getUrl());
			$_SESSION["REALTIMEENABLED"] = $data == "" || strlen($data) < 10 ? false : true;
			return $_SESSION["REALTIMEENABLED"];
			*/
	}

	/*static public function isNewDataAvaileble() {
		$data = file_get_contents(self::getUrl());

		preg_match("#<meta name=\"collection_time\" content=\"(.*?)\" />#imsU", $data, $out);
		$newCollectionTime = isset($out[1]) ? $out[1] : '';

		$realTime = RealTime::getInstance();
		if ($newCollectionTime > $realTime->getLastUpdated(true)) {
		return true;
		}

		return false;
		}*/


	/**
	 * @return RealTime
	 */
	static public function getInstance($fresh = false, $isLM= false) {
		Logger::CRITICAL('getInstance', "getInstance ('{$fresh}')");
		if (is_null(self::$instance)) {

			if (!isset($_SESSION["REALTIME"])) {
				$_SESSION["REALTIME"] = array();
				$_SESSION["REALTIME"]["DATA"] = null;
				$_SESSION["REALTIME"]["DATATIMESTAMP"] = date("U");
				$_SESSION["REALTIME"]["SELECTION"] = array();
				$_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"] = array();
				$_SESSION["REALTIME"]["SELECTION"]["FEATURES"] = array();
				$_SESSION["REALTIME"]["SELECTION"]["USERS"] = array();
				$_SESSION["REALTIME"]["DATA"] = new RealTime();
				
				if(SystemSettings::get(SystemSettings::SERVER_NS . "licensemonitor.overlay.mapping.data") > 0 && !isset($_SESSION["USERDATA"])  && $isLM){
					Realtime::populateUserData($_SESSION["REALTIME"]["DATA"]);
				}
			}
			
			

		}

		// If content is to old (4 min), reload it...
		if ($fresh || ($_SESSION["REALTIME"]["DATATIMESTAMP"] + 240) < date("U")) {
			$_SESSION["REALTIME"]["DATA"] = new RealTime();
			if(SystemSettings::get(SystemSettings::SERVER_NS . "licensemonitor.overlay.mapping.data") > 0 && $isLM){
				Realtime::populateUserData($_SESSION["REALTIME"]["DATA"]);
			}
			$_SESSION["REALTIME"]["DATATIMESTAMP"] = date("U");
		}
		
		if(!$_SESSION["ERROR_MESSAGE"] && strlen($_SESSION["ERROR_MESSAGE"]) <= 0){
			Realtime::checkMessages();
		}
		
		// Get instance from cache..
		self::$instance = $_SESSION["REALTIME"]["DATA"];
		return self::$instance;
	}


	static public function populateUserData($rt){
		//$realTime = $_SESSION["REALTIME"]["DATA"];
		$_SESSION["USERDATA"]["SELECTION"] = array();
		$userData = array();
		
		
		// Get realtime data
		foreach($rt->getUsers() as $userd){
			$sql = "SELECT name,value FROM vData where id='" . $userd->name ."' AND length(value) > 0 GROUP BY name";
			$_SESSION["USERDATA"]["SELECTION"][$userd->name] = array();
			$_SESSION["USERDATA"]["SELECTION"][$userd->name] = MappingDatabaseConnector::query($sql);
			
			$hasGN=false;
			$coun=0;
			foreach($_SESSION["USERDATA"]["SELECTION"][$userd->name] as $usern){
				$uDataSize=sizeOf($_SESSION["USERDATA"]["SELECTION"][$userd->name]);
				$_SESSION["USERDATA"]["SELECTION"][$userd->name][$coun]["value"]=Realtime::applyFilter($usern);

				if ($usern['name']=="cn") {
					$_SESSION["USERDATA"]["SELECTION"][$userd->name][$uDataSize]['name'] = 'displayName';
					$_SESSION["USERDATA"]["SELECTION"][$userd->name][$uDataSize]['value'] = $usern['value'];
				}
				elseif($usern['name']=="givenName") {
					$dispNameDex=$uDataSize;
					$_SESSION["USERDATA"]["SELECTION"][$userd->name][$uDataSize]['name'] = 'displayName';
					$_SESSION["USERDATA"]["SELECTION"][$userd->name][$uDataSize]['value'] = $usern['value'];
					$hasGN=true;
				}
				if ($usern['name']=="sn"&&$hasGN) $_SESSION["USERDATA"]["SELECTION"][$userd->name][$dispNameDex]['value'] = $usern['value'].", ". $_SESSION["USERDATA"]["SELECTION"][$userd->name][$dispNameDex]['value'];
				$coun++;
			}
		}
		//return $userData;
	}
	
	static private function applyFilter($uDataRow){
		$filtered;
		Logger::DEBUG('applyFilter', "Applying Filter for ('{$uDataRow['name']}') = ('{$uDataRow['value']}')");
		$mapFilter = new MapFilter('/var/opt/openit/etc/mapfilters.cfg');
		$filtered=$uDataRow['value'];
		foreach($mapFilter->filters as $frules){		
			if ($frules['fieldType']=='dest'){
				switch($frules['action']){
					case 'trim':
						if(preg_match("[".$frules['field']."]",$uDataRow['name'],$matches)){
							$filtered= trim($filtered);

						}
						break;
					
					case 'setcase':
						if(preg_match("[".$frules['field']."]",$uDataRow['name'],$matches)){
							if($frules['param1']=='lower'){
								$filtered= strtolower($filtered);
							}elseif($frules['param1']=='upper'){
								$filtered= strtoupper($filtered);
							}elseif($frules['param1']=='capitalized'){
								$filtered= ucfirst($filtered);
							}
						}
						break;
					
					case 'translate':
						if(preg_match("[".$frules['field']."]",$uDataRow['name'],$matches)){
							$x=0;
							for ($i=0;$i<strlen($frules['param1']);$i++){
								if(substr($frules['param2'],$x,1)=="\\"){
									$repl=substr($frules['param2'],$x,2);
								}else{
									$repl=substr($frules['param2'],$x,1);
								}
								
								if(substr($frules['param1'],$i,1)=="\\"){
									$filtered=str_replace(substr($frules['param1'],$i,2),$repl,$filtered);
									$i++;
								}else{
									$filtered=str_replace(substr($frules['param1'],$i,1),$repl,$filtered);
								}
								$x++;
								if($x>strlen($frules['param2'])-1)$x=strlen($frules['param2'])-1;
							}
						}
						break;
							
					case 'map':
						if($frules['field']==$uDataRow['name']){
							if(preg_match("[^".$frules['param1']."$]",$filtered,$matches)){
								$filtered= $frules['param2'];
							}
						}
						break;
					case 'imap':
						if($frules['field']==$uDataRow['name']){
							if(preg_match("[^".$frules['param1']."$]i",$filtered,$matches)){
								$filtered= $frules['param2'];
							}
						}
						break;
				}
			}
		}
		return $filtered;

	}
	
	static public function clearSelections() {
		$_SESSION["REALTIME"]["SELECTION"] = array();
		$_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"] = array();
		$_SESSION["REALTIME"]["SELECTION"]["FEATURES"] = array();
		$_SESSION["REALTIME"]["SELECTION"]["USERS"] = array();
	}

	public function setSelections($type, $values) {
		$_SESSION["REALTIME"]["SELECTION"][strtoupper($type) . "S"] = $values;
	}

	public function getSelections($type) {
		return $_SESSION["REALTIME"]["SELECTION"][strtoupper($type) . "S"];
	}


	public function addtSelection($type, $selection) {
		$_SESSION["REALTIME"]["SELECTION"][strtoupper($type) . "S"][] = $selection;
	}

	private function readXML_OLD($data) {

		// Check if we have data..
		if ($data == "" || strlen($data) < 10) return;
		$featureRenameMap = $_SESSION['featureRenameMap'];
		$productRenameMap = $_SESSION['productRenameMap'];
		$xml = XMLParser::xmlToObj($data);

		$realTimeMap = array();

		// Check if the XML is empty
		if (is_null($xml->meta[0])) {
			throw new Exception("Not valid XML!");
		}

		if (!is_null($xml->error)) {
			$_SESSION["ERROR_MESSAGE"] = $xml->error->message;
		}

		$this->collectionTime = $xml->meta[0]->get("content");
		$filter = new Filter("/var/opt/openit/etc/exclude-mapping");
		foreach ($xml->vendorlicenses[0]->children() as $child) {
			if ($child->getName() == "vendorlicense") {

				// Create product object
				$product = new RealTimeProduct();
				$product->name = $child->get("name");
				if(count($productRenameMap) > 0 ){
					$product->name= ($productRenameMap[$product->name] != NULL || $productRenameMap[$product->name] != "" ) ? $productRenameMap[$product->name] : $product->name;
				}
				$product->features = array();
				$product->server = $child->servers[0]->server[0]->get("name");
				$product->licenseFiles = $this->formatLicense($child->license-file);

				foreach ($child->daemons[0]->daemon[0]->features[0]->children() as $childId => $featureChild) {
					if ($featureChild->getName() == "feature") {

						// Create feature object
						$feature = new RealTimeFeature();
						$feature->name = $featureChild->get("name");
						if(count($featureRenameMap) > 0 ){
							$feature->name = ($featureRenameMap[$feature->name] != NULL || $featureRenameMap[$feature->name] != "" ) ? $featureRenameMap[$feature->name] : $feature->name;
						}
						$feature->licensesTotal = $featureChild->get("licenses");
						$feature->expiry = $featureChild->get("expires");
						$feature->version = $featureChild->get("version");
						$feature->license = $product->getLicenseFile($product->name,$feature->name);
						$feature->productName = $product->name;
						$feature->licenseServer = $product->server;

						$users = array();

						foreach ($featureChild->online[0]->children() as $onlineEntry) {
							$user = new RealTimeUser();
							$user->name = $onlineEntry->user[0]->getText();
							$user->display = $onlineEntry->display[0]->getText();
							$user->host = $onlineEntry->host[0]->getText();
							$user->start = $onlineEntry->start[0]->getText();
							$user->count = $onlineEntry->count[0]->getText();
							
							$user->licensesTotal = $featureChild->get("licenses");	// Static

							$users[] = $user;
						}

						// Add users to feature
						$feature->users = $users;

						// Add feature to product
						if(!is_null( $filter) && !($filter->isExcluded($product->name,$feature->version))){
							$product->addFeature($feature);
						}
					}
				}
				$realTimeMap[] = $product;
			}
		}
		return $realTimeMap;
	}

	private function readXML($data) {

		$featureRenameMap = $_SESSION['featureRenameMap'];
		$productRenameMap = $_SESSION['productRenameMap'];
		// Get product mapping name
		$productMapping= null;
		$selection = new ReportSelection();
		if ($selection->get(RealTime::GUI_PRODUCT_MAPPING) == 1) {
			$productMapping = new ProductMapping('/var/opt/openit/etc/product-mapping');
		}

		// Check if we have data..
		if ($data == "" || strlen($data) < 10) return;
		$charset = ServerConfig::getCharset();
		Logger::Debug('RealTime::readXML()', "charset ('{$charset}').");
		$xml = new SimpleXMLElement($data);
		//var_dump($xml);

		$realTimeMap = array();

		if (!is_null($xml->error)) {
			$_SESSION["ERROR_MESSAGE"] = $xml->error->message;
		}

		// Check if the XML is empty
		if (is_null($xml->meta)) {
			throw new Exception("Not valid XML!");
		}

		$this->collectionTime = (string) $xml->meta["content"];

		$reUse = false;
		$filter = new Filter("/var/opt/openit/etc/exclude-mapping");
		foreach ($xml->vendorlicenses as $vendorlicenses) {
			foreach ($vendorlicenses->vendorlicense as $vendorlicense) {

				$feature = null;
				$product = null;
				$orignialProduct = null;

				// Find product name
				
				$productName = (string) $vendorlicense['name'];
				if(count($productRenameMap) > 0 ){
					$productName = ($productRenameMap[$productName] != NULL || $productRenameMap[$productName] != "" ) ? $productRenameMap[$productName] : $productName;
				}
				//$orignialProductName = (string) $vendorlicense['name'];

				// Check for mapping
				if (!is_null($productMapping) && !is_null($productMapping->getParent($productName))) {
					$productName = $productMapping->getParent($productName);
				}

				$reUse = false;

				// Do we have this product in our map?
				foreach ($realTimeMap as $product) {
					if ($product->name == $productName) {
						$reUse = true;
						Logger::Debug('RealTime::readXML()', "Found existing product name ('{$productName}') in map, reusing this object.");
						break;
					}
					$product = null;
				}

				// Create product object if its empty
				if (is_null($product)) {
					Logger::Debug('RealTime::readXML()', "Creating new product name ('{$productName}') in map.");
					$product = new RealTimeProduct();
					$product->name = $productName;
					$product->polltime = (int) $vendorlicense->polltime[0];
					$product->features = array();
					$product->server = (string) $vendorlicense->servers[0]->server[0]['name'];
					$product->serverport = (string) $vendorlicense->servers[0]->server[0]['port'];
					$product->daemon = (string) $vendorlicense->daemons[0]->daemon[0]['name'];
					$product->daemonversion = (string) $vendorlicense->daemons[0]->daemon[0]['version'];
					$license = (string) $vendorlicense->licensefile[0];
					$product->licenseFiles = $this->formatLicense($license);

				}

				// Create the original as well
				/*
				 if ($productName != $orignialProductName) {
					$orignialProduct = new RealTimeProduct();
					/ **
					* To sort the products correctly, we have a hack..
					* We add the new global mapping name first, then adding ## -- and the original name..
					* Then when we display it, we remove front name when we have ## -- in the name string..
					* /
					$orignialProduct->name = $productName . '## -- ' . $orignialProductName;
					$orignialProduct->features = array();
					}
					*/
					
				foreach ($vendorlicense->daemons->daemon->features->feature as $featureChild) {


					// Create feature object
					$feature = new RealTimeFeature();
					$feature->name = (string) $featureChild['name'];
					if(count($featureRenameMap) > 0 ){
						$feature->name = ($featureRenameMap[$feature->name] != NULL || $featureRenameMap[$feature->name] != "" ) ? $featureRenameMap[$feature->name] : $feature->name;
					}
					$feature->licensesTotal = (string) $featureChild['licenses'];
					$feature->expiry = (string) $featureChild['expires'];
					$feature->version = (string) $featureChild['version'];
					$feature->license = $product->getLicenseFile($product->name,$feature->name);
					$feature->productName = $product->name;
					$feature->licenseServer = $product->server;

					
					Logger::Debug('RealTime::readXML()', "Adding new feature name ('{$feature->name}') in map.");
					$users = array();

					foreach ($featureChild->online->entry as $onlineEntry) {
						$user = new RealTimeUser();
						$user->name = (string) $onlineEntry->user;
						/*if($overlayFlag > 0){
							$sql = "SELECT value FROM vData where id='" . $user->name ."' AND length(value) > 0 AND name='cn'";
							$result = MappingDatabaseConnector::query($sql);
							$displayName = $result[0]['value'];
							$user->name = strlen($displayName) > 0 ? $displayName : $user->name;
						}*/
						$user->display = (string) $onlineEntry->display;
						$user->host = (string) $onlineEntry->host;
						$user->start = (string) $onlineEntry->start;
						$user->count = (string) $onlineEntry->count;

						$user->licensesTotal = (string) $featureChild['licenses']; // STATIC!
							
						$users[] = $user;
					}

					// Add users to feature
					$feature->users = $users;

					// And the original product as well
					/* if ($productName != $orignialProductName) {
						Logger::Debug(__CLASS__, "Adding {$orignialProduct->name} {$feature->name} to orginal with {$feature->licensesTotal} licenses total..");
						$orignialProduct->addFeature(clone $feature); // Make sure its a clone
						}*/

					// Add feature to product
					if(!is_null( $filter) && !($filter->isExcluded($product->name,$feature->version))){
						$product->addFeature($feature);
					}

				}

				// We dont add it again, if we are reusing an existing object..
				if (!$reUse) {
					$realTimeMap[] = $product;
				}

				/*if ($productName != $orignialProductName) {
					$realTimeMap[] = $orignialProduct;
					}*/
			}
		}
		//var_dump($realTimeMap);
		return $realTimeMap;
	}
		
	static public function checkMessages() {
		$data = file_get_contents(self::getUrl());
		// TODO: SHOULD BE FIXED IN BACKEND!
		$data = str_replace("&", "&amp;", $data);

		// Check if we have data..
		if ($data == "" || strlen($data) < 10) return;

		if (class_exists('SimpleXMLElement', false)) {
			$xml = new SimpleXMLElement($data);
			
		}else {
			$xml = XMLParser::xmlToObj($data);
		}
		if(!is_null($xml->error->message)){
			$_SESSION["ERROR_MESSAGE"] = $xml->error->message;
		}else if(count($xml->vendorlicenses)==0 || count($xml->vendorlicenses[0]) == 0){
			$_SESSION["ERROR_MESSAGE"] = "We could not locate any License Snapshot Status data.<br /><br />If you have just installed this application, please try again after approximately 1 hour.";
		}
	}


	public function getLastUpdated($asEpoch = false) {
		if ($asEpoch) return $this->collectionTime;
		return date("Y-m-d H:i", $this->collectionTime);
	}


	public function getRealTimeMap() {
		return $this->realTimeMap;
	}

	public function getProducts($selectedOnly = false, $sortBy = null, $sortAsc = null) {
		if ($selectedOnly && count($_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"]) != 0) {
			$products = array();
			foreach ($this->genericFilter(self::PRODUCT, $sortBy, $sortAsc) as $product) {
				if (in_array($product->name, $_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"])) {
					$products[] = $product;
				}
			}
			return $products;
		}
		return $this->genericFilter(self::PRODUCT, $sortBy, $sortAsc);
	}

	public function getFeatures($selectedOnly = false, $sortBy = null, $sortAsc = null) {
		if ($selectedOnly && count($_SESSION["REALTIME"]["SELECTION"]["FEATURES"]) != 0) {
			$features = array();
			foreach ($this->genericFilter(self::FEATURE, $sortBy, $sortAsc) as $feature) {
				if (in_array($feature->name, $_SESSION["REALTIME"]["SELECTION"]["FEATURES"])) {
					$features[] = $feature;
				}
			}
			return $features;
		}
		return $this->genericFilter(self::FEATURE, $sortBy, $sortAsc);
	}

	public function getUsers($selectedOnly = false) {
		if ($selectedOnly && count($_SESSION["REALTIME"]["SELECTION"]["USERS"]) != 0) {
			$entries = array();
			foreach ($this->genericFilter(self::USER) as $entry) {
				if (in_array($entry->name, $_SESSION["REALTIME"]["SELECTION"]["USERS"])) {
					$entries[] = $entry;
				}
			}
			return $entries;
		}
		return $this->genericFilter(self::USER);
	}

	private function genericFilter($type, $sortBy = null, $sortAsc = null) {
		$objectKey = array();
		$objects = array();

		$filter = new Filter("/var/opt/openit/etc/exclude-mapping");
				
		// Loop thru lists..
		foreach ($this->realTimeMap as $product) {
			// Product
			if ($type != self::PRODUCT && count($_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"]) != 0) {
				if (!in_array($product->name, $_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"])) {
					continue;
				}
			}

			// If no users, we need to add products
			/*
			 if (count($product->features) == 0 && $type == self::PRODUCT) {
				if (!isset($objectKey[$product->name])) {
				$objects[] = $product;
				$objectKey[$product->name] = 1;
				}
				continue;
				}
				*/
			foreach ($product->features as $feature) {

				// Feature
				if ($type != self::FEATURE && count($_SESSION["REALTIME"]["SELECTION"]["FEATURES"]) != 0) {
					if (!in_array($feature->name, $_SESSION["REALTIME"]["SELECTION"]["FEATURES"])) {
						continue;
					}
				}


				// If no users, we need to add features and products
				/*
				 if (count($feature->users) == 0 && $type != self::USER) {
					if ($type == self::PRODUCT) {
					if (!isset($objectKey[$product->name])) {
					$objects[] = $product;
					$objectKey[$product->name] = 1;
					}
					continue 2;
					} else if ($type == self::FEATURE) {
					if (!isset($objectKey[$feature->name])) {
					$objects[] = $feature;
					$objectKey[$feature->name] = 1;
					}
					continue;
					}
					}


					* */
				// If we have no users
				if (count($feature->users) == 0 || $type != self::USER) {

					// Add objects
					if ($type == self::PRODUCT) {
						if (!isset($objectKey[$product->name])) {
							$objects[] = $product;
							$objectKey[$product->name] = 1;
						}
						continue 2;
					} else

					if ($type == self::FEATURE) {
						if (!isset($objectKey[$feature->name])) {
							$objects[] = clone $feature;
							$objectKey[$feature->name] = 1;

							// Feature exist in list, merge list
						}else {
							foreach ($objects as $objectEntry) {
								if ($objectEntry->name == $feature->name) {
									$objectEntry->licensesTotal += $feature->licensesTotal;
									$objectEntry->users = array_merge($objectEntry->users, $feature->users);
								}
							}
						}
						//}
						continue 1;
					}

				} else {
					foreach ($feature->users as $user) {

						// Add if there is no user selections..
						if ($type != self::USER && count($_SESSION["REALTIME"]["SELECTION"]["USERS"]) != 0) {
							if (!in_array($user->name, $_SESSION["REALTIME"]["SELECTION"]["USERS"])) {
								continue;
							}
						}

						if ($type == self::USER) {
							if (!isset($objectKey[$user->name])) {
								/*if(SystemSettings::get(SystemSettings::SERVER_NS . "licensemonitor.overlay.mapping.data", 0) > 0){
									$sql = "SELECT name,value FROM vData where id='" . $user->name ."' AND length(value) > 0 AND value GROUP BY name";
									$user->setUserData(MappingDatabaseConnector::query($sql));
								}*/
								$objects[] = $user;
								$objectKey[$user->name] = 1;
							}
						}
					}
					//}

				}
			}
		}


		// Sorting
		if (!is_null($sortBy)) {

			switch (strtoupper($sortBy)) {
				
				case "PRODUCT":
					// sort by feature name
					usort($objects, "RealTime_sort_product");
					break;
				
				case "VERSION":
					// sort by feature name
					usort($objects, "RealTime_sort_version");
					break;
				
				case "EXPIRY":
					// sort by feature name
					usort($objects, "RealTime_sort_expiry");
					break;
									
				case "TOTAL":
					// sort by license totalt
					usort($objects, "RealTime_sort_total");
					break;

				case "USED":
					// sort by used licenses
					usort($objects, "RealTime_sort_used");
					break;

				case "UNUSED":
					// sort by un used licenses
					usort($objects, "RealTime_sort_unused");
					break;

				case "TIME":
					// sort by checkout time
					usort($objects, "RealTime_sort_time");
					break;

				default:
					// sort by name
					usort($objects, "RealTime_sort");
					break;
			}
		}

		// default sorting - by name
		else {
			usort($objects, "RealTime_sort");
		}
		
		// Sort direction
		if (!is_null($sortAsc) && $sortAsc) {
			$objects = array_reverse($objects);
		}
		return $objects;
	}

	public function hasSelectedFeatures() {
		foreach ($this->getFeatures() as $feature) {
			if ($feature->isSelected()) return true;
		}
		return false;
	}

}

function RealTime_sort($a, $b) {
	return strcasecmp($a->name, $b->name); // by name - works on all
}

// on users
function RealTime_sort_host($a, $b) {
	return strcasecmp($a->host, $b->host);
}

function RealTime_sort_count($a, $b) {
	return $a->count == $b->count ? 0 : ($a->count < $b->count ? -1 : 1);
}

// on product and feature

function Realtime_sort_expiry($a, $b) {
	if (($a instanceof RealTimeFeature && $b instanceof RealTimeFeature)) {
		return $a->getExpiryEpoch() == $b->getExpiryEpoch() ? 0 : ($a->getExpiryEpoch() < $b->getExpiryEpoch() ? -1 : 1); // by expiry - only in feature view
	}
	return strcasecmp($a->name, $b->name); // by name - works on all
}

function Realtime_sort_version($a, $b) {
	if (($a instanceof RealTimeFeature && $b instanceof RealTimeFeature)) {
		return $a->version == $b->version ? 0 : ($a->version < $b->version ? -1 : 1); // by feature version - only in feature view
	}
	return strcasecmp($a->name, $b->name); // by name - works on all
}

function Realtime_sort_product($a, $b) {
	if (($a instanceof RealTimeFeature && $b instanceof RealTimeFeature)) {
		return $a->getProductName() == $b->getProductName() ? 0 : ($a->getProductName() < $b->getProductName() ? -1 : 1); // by product name - only in feature view
	}
	return strcasecmp($a->name, $b->name); // by name - works on all
}

function RealTime_sort_total($a, $b) {
	if (($a instanceof RealTimeFeature && $b instanceof RealTimeFeature) || ($a instanceof RealTimeProduct && $b instanceof RealTimeProduct)) {
		return $a->getRealTimeTotaltLicenses() == $b->getRealTimeTotaltLicenses() ? 0 : ($a->getRealTimeTotaltLicenses() < $b->getRealTimeTotaltLicenses() ? -1 : 1); // by total licenses - only in feature view
	}
	return strcasecmp($a->name, $b->name); // by name - works on all
}

function RealTime_sort_used($a, $b) {
	if (($a instanceof RealTimeFeature && $b instanceof RealTimeFeature) || ($a instanceof RealTimeProduct && $b instanceof RealTimeProduct)) {
		return $a->getRealTimeLicenseUsage() == $b->getRealTimeLicenseUsage() ? 0 : ($a->getRealTimeLicenseUsage() < $b->getRealTimeLicenseUsage() ? -1 : 1); // by used licenses - only in feature view
	}
	return strcasecmp($a->name, $b->name); // by name - works on all
}

function RealTime_sort_unused($a, $b) {
	if (($a instanceof RealTimeFeature && $b instanceof RealTimeFeature) || ($a instanceof RealTimeProduct && $b instanceof RealTimeProduct)) {
		return ($a->getRealTimeTotaltLicenses() - $a->getRealTimeLicenseUsage()) == ($b->getRealTimeTotaltLicenses() - $b->getRealTimeLicenseUsage()) ? 0 : (($a->getRealTimeTotaltLicenses() - $a->getRealTimeLicenseUsage()) < ($b->getRealTimeTotaltLicenses() - $b->getRealTimeLicenseUsage()) ? -1 : 1); // by un-used licenses - only in feature view
	}
	return strcasecmp($a->name, $b->name); // by name - works on all
}

function RealTime_sort_time($a, $b) {
	if (($a instanceof RealTimeFeature && $b instanceof RealTimeFeature) || ($a instanceof RealTimeProduct && $b instanceof RealTimeProduct)) {
		return $a->getFirstChecoutTime() == $b->getFirstChecoutTime() ? 0 : ($a->getFirstChecoutTime() < $b->getFirstChecoutTime() ? -1 : 1); // by used licenses - only in feature view
	}

	else if ($a instanceof RealTimeUser  && $b instanceof RealTimeUser) {
		return $a->start == $b->start ? 0 : ($a->start < $b->start ? -1 : 1); // by used licenses - only in feature view
	}
	return strcasecmp($a->name, $b->name); // by name - works on all
}

function RealTime_sort_total2($a, $b) {
	//Logger::Debug("GLENN", $a->licensesTotal . " <==> " .$b->licensesTotal);
	//return strcasecmp($a->name, $b->name); // by name - works on all
	//return $a->licensesTotal == $b->licensesTotal ? 0 : ($a->licensesTotal < $b->licensesTotal ? -1 : 1); // by total licenses - only in feature view
	//return count($a->users) == count($b->users) ? 0 : (count($a->users) < count($b->users) ? -1 : 1); // by used licenses - only in feature view
	//return ($a->licensesTotal - count($a->users)) == ($b->licensesTotal - count($b->users)) ? 0 : (($a->licensesTotal - count($a->users)) < ($b->licensesTotal - count($b->users)) ? -1 : 1); // by un-used licenses - only in feature view

	// Sort by cheackout (start) time

	if ($a instanceof RealTimeFeature && $b instanceof RealTimeFeature) {
		return $a->getFirstChecoutTime() == $b->getFirstChecoutTime() ? 0 : ($a->getFirstChecoutTime() < $b->getFirstChecoutTime() ? -1 : 1); // by used licenses - only in feature view
	}

	if ($a instanceof RealTimeUser && $b instanceof RealTimeUser) {
		return $a->start == $b->start ? 0 : ($a->start < $b->start ? -1 : 1); // by used licenses - only in feature view
	}

	return strcasecmp($a->name, $b->name); // by name - works on all
}

class RealTimeProduct {
	public $name;
	public $polltime;
	public $features;
	public $server;
	public $serverport;
	public $daemon;
	public $licenseFiles;
	public $daemonversion;

	public function getMinutesSinceUpdate() {
		//		return $this->polltime;
		//return floor((date("U") - $this->polltime) / 60);
		return floor((time() - $this->polltime) / 60);
	}

	public function getLastSeen() {
		//return (time() + $offset) - $this->polltime;
		return time()- $this->polltime;
	}

	public function getLastSeenFormat() {
		
		//if((time() + $offset) > $this->polltime){
		if(time() > $this->polltime){
			//$timeSpent = (time() + $offset) - $this->polltime;
			$timeSpent = time() - $this->polltime;
			$days = floor($timeSpent / 86400);
			$hours = floor(($timeSpent - ($days * 86400)) / 3600);
			$minutes = floor(($timeSpent - ($days * 86400) - ($hours * 3600)) / 60);
			return ($days != 0 ? $days . " day(s), " : "") . ($hours != 0 ? $hours . " hour(s), " : "") . $minutes . " minute(s)";
		} else{
			return 'N/A';
		}
	}

	public function isSelected() {
		return in_array($this->name, $_SESSION["REALTIME"]["SELECTION"]["PRODUCTS"]);
	}

	public function addFeature(RealTimeFeature $feature) {

		$isNew = true;

		// Check if we have the feature allready
		foreach ($this->features as &$eFeature) {
			if ($eFeature->name == $feature->name && $eFeature->version == $feature->version) {
				Logger::Debug(__CLASS__, "Found existing feature ({$feature->name}) in product {$this->name}...");
				$isNew = false;
				break;
			}
		}

		// Add if feature is new
		if ($isNew) {
			$this->features[] = $feature;
			return;
		}

		// If feature is not new, we need to merge..
		//Logger::Debug(__CLASS__, "Old {$feature->name} is {$eFeature->licensesTotal}");
		//Logger::Debug(__CLASS__, "New {$feature->name} is {$feature->licensesTotal}");
		$eFeature->licensesTotal += $feature->licensesTotal;
		$eFeature->users = array_merge($eFeature->users, $feature->users);
		Logger::Debug(__CLASS__, "New total for {$feature->name} is {$eFeature->licensesTotal}");

	}

	public function getRealTimeLicenseUsage($selected = false) {
		$count = 0;
		foreach ($this->features as $feature) {
			if (!$selected || count($_SESSION["REALTIME"]["SELECTION"]["FEATURES"]) == 0) {
				$count += $feature->getRealTimeLicenseUsage($selected);
			} else {
				if (in_array($feature->name, $_SESSION["REALTIME"]["SELECTION"]["FEATURES"])) {
					$count += $feature->getRealTimeLicenseUsage($selected);
				}
			}
		}
		return $count;
	}

	public function getFeatures(){
		return $this->features;
	}

	public function getUsers() {
		$users = array();
		foreach ($this->features as $feature) {
			$users = array_merge($feature->getUsers(), $users);
		}
		return $users;
	}

	public function getFirstChecoutTime() {
		if (count($this->getUsers()) == 0) return null;
		$first = null;
		foreach ($this->getUsers() as $user) {
			if ($first == null || $user->start < $first) {
				$first = $user->start;
			}
		}
		return $first;
	}

	public function getRealTimeTotaltLicenses() {
		$total = 0;
		foreach ($this->features as $feature) {
			$total += $feature->getRealTimeTotaltLicenses();
		}
		return $total;
	}

	function longestSubstr($str,$toCompare)
	{
		$len1 = strlen($str);
		$len2 = strlen($toCompare);
		if ($len1==0 || $len2==0)
		return 0;
		$maxLen = 0;

		$compareTable = array();

		for($i=0; $i<$len1; $i++){
			$compareTable[$i] = array();
			for($j=0; $j<$len2; $j++){
				$compareTable[$i][$j] = 0;
			}
		}

		for ($m = 0; $m < $len1; $m++)
		{
			for ($n = 0; $n < $len2; $n++)
			{
				$compareTable[$m][$n] = ($str[$m] != $toCompare[$n]) ? 0 : ((($m == 0) || ($n == 0)) ? 1 : $compareTable[$m - 1][$n - 1] + 1);
				$maxLen = ($compareTable[$m][$n] > $maxLen) ? $compareTable[$m][$n] : $maxLen;
			}
		}
		return $maxLen;
	}

	//Get the specific license file for a feature
	function getLicenseIndex($strarr,$feature_name){

		/*if(is_array() && count($strarr) == 1){
			return 0;
			}else if(is_array() && count($strarr) > 1){
			*/
		$match_id = 0;
		$previous_max = 0;
		for($i = 0; $i < count($strarr); $i++){
			$license_file = $strarr[$i];

			$tmp = $this->longestSubstr($license_file,$feature_name);
			if($previous_max < $tmp){
				$previous_max = $tmp;
				$match_id = $i;
			}
		}

		return $match_id;
		/*}else{
			return null;
			}*/
	}

	public function getLicenseFile($product_name,$feature_name){
		$license = $this->licenseFiles;
		$license = str_replace(":\\", "REPLACE", $license);
		$strarr = split(":",$license);
		$strarr = str_replace("REPLACE", ":\\", $strarr);

		$path = $strarr[$this->getLicenseIndex($strarr,$feature_name)];

		if (strpos($path, '/') !== false) $basename = end(explode('/', $path));
		elseif (strpos($path, '\\') !== false) $basename = end(explode('\\', $path));
		else return false;
		if (empty($basename)) return false;

		$dirname = substr($path, 0, strlen($path) - strlen($basename) - 1);

		if (strpos($basename, '.') !== false)
		{
			$extension = end(explode('.', $path));
			$filename = substr($basename, 0, strlen($basename) - strlen($extension) - 1);
		}
		else
		{
			$extension = '';
			$filename = $basename;
		}

		return array(
      	'dirname' => $dirname, 
      	'basename' => $basename, 
      	'extension' => $extension, 
      	'filename' => $filename 
		);
	}

}

class RealTimeFeature {
	public $name;
	public $users;
	public $licensesTotal;
	public $expiry;
	public $version;
	public $license;
	public $status;
	public $productName;
	public $licenseServer;

	public function isSelected() {
		return in_array($this->name, $_SESSION["REALTIME"]["SELECTION"]["FEATURES"]);
	}
	
	public function getLicenseFile(){
		return $this->license;
	}

	public function getProductName(){
		return $this->productName;
	}
	
	public function getLicenseServer(){
		return $this->licenseServer;
	}
	
	public function getUsers($sortBy = false, $sortAsc = false) {

		$users = $this->users;

		if ($sortBy) {
			switch ($sortBy) {
				case "name":
					usort($users, "RealTime_sort"); // by name
					break;

				case "host":
					usort($users, "RealTime_sort_host"); // by host
					break;

				case "time":
					usort($users, "RealTime_sort_time"); // by checkouttime
					break;

				case "count":
					usort($users, "RealTime_sort_count"); // by licenses in use
					break;
			}

			if (!is_null($sortAsc) && $sortAsc) {
				$users = array_reverse($users);
			}
		}

		return $users;
	}

	public function getRealTimeLicenseUsage($selected = false) {
		$count = 0;
		foreach ($this->users as $user) {
			if (!$selected || count($_SESSION["REALTIME"]["SELECTION"]["USERS"]) == 0) {
				$count += $user->count;
			} else {
				if (in_array($user->name, $_SESSION["REALTIME"]["SELECTION"]["USERS"])) {
					$count += $user->count;
				}
			}
		}
		return $count;
	}

	public function getFirstChecoutTime() {
		if (count($this->users) == 0) return null;
		$first = null;
		foreach ($this->users as $user) {
			if ($first == null || $user->start < $first) {
				$first = $user->start;
			}
		}
		return $first;
	}

	public function getRealTimeTotaltLicenses() {
		return $this->licensesTotal;
	}

	public function getFeatures(){
		return $this->features;
	}

	public function getExpiryFormat($offset = 0) {
		if((time() + $offset) < $this->expiry){
			$timeToExpire = $this->expiry - (time() + $offset);
		} else{
			$timeToExpire = (time() + $offset) - $this->expiry;
		}
		//$timeToExpire = $this->getExpiryEpoch($offset);
		$days = floor($timeToExpire / 86400);
		$hours = floor(($timeToExpire - ($days * 86400)) / 3600);
		$minutes = floor(($timeToExpire - ($days * 86400) - ($hours * 3600)) / 60);
		return ($days != 0 ? $days . " day(s), " : "") . ($hours != 0 ? $hours . " hour(s), " : "") . $minutes . " minute(s)";
	}

	//Get the expiry in number of seconds
	public function getExpiryEpoch($offset = 0){
		return $this->expiry - (time() + $offset);
	}

	//Get the expiry in number of hours
	public function getExpiry($offset = 0){
		if((time() + $offset) < $this->expiry){
			$timeToExpire = $this->expiry - (time() + $offset);
			$hours = floor($timeToExpire / 3600);
			return $hours;
		} else{
			return 'N/A';
		}
	}

	public function getExpiryValue(){
		return $this->expiry;
	}

}

class RealTimeUser {
	public $name;
	public $display;
	public $host;
	public $start;
	public $count;
	public $userData = array();

	public $licensesTotal;

	public function isSelected() {
		return in_array($this->name, $_SESSION["REALTIME"]["SELECTION"]["USERS"]);
	}

	public function getRealTimeLicenseUsage() {
		return $this->count;
	}

	public function getRealTimeTotaltLicenses() {
		return $this->licensesTotal;
	}
	
	public function setUserData($userdata){
		$this->userData = $userdata;
	}
	
	public function getUserData(){
		return $this->userData;
	}
}


?>